#!/usr/bin/python
#################################################################################################################################
""" This script converts Pilatus/TIFF files to NeXus (http://www.nexusformat.org/)

tiff2nexus expects a directory containing TIFF files written by a Pilatus detector.
The TIFF files have to be named in the following way:

   anyname_#.tif
   
where # is an integer number with one or more digits. Other files in the input
directory will be ignored. tiff2nexus writes the names of the ignored files to
standard output. The script can create NeXus/HDF5 (default) or NeXus/XML. All
Pilatus/TIFF files are united to one NeXus file.

tiff2nexus writes NeXus groups and data in the following order:

entry:NXentry
    title = dirname
    start_time = timestamp of the first image
    source:NXsource
        type = "Synchrotron"
        name = "PetraIII"
    instrument:NXinstrument
        name = "P03 GISAXS"
        detector:NXdetector
            aequatorial_angle = 0.
            azimuthal_angle = 0.
            beam_center_x = n_xPixel * x_pixel_size * 0.5
                @unit
            beam_center_y = n_yPixel * y_pixel_size * 0.5
                @unit
            data (3-dim array of image data)
                @axes = "image:x:y" (meaning of indices)
            distance = 0.
            polar_angle = 0.
            x_pixel_size
                @unit
            y_pixel_size
                @unit
            image_header:NXnote
                type = "text/plain"
                description = those parts of header text that isn't written elsewhere
        monochromator:NXmonochromator
            wavelength = 1.
            wavelength_spread = 0.
    sample:NXsample
        aequatorial_angle = 0.
        rotation_angle = 0.
        name = dirname
        image_timestamp = timestamp for each TIFF file
        exposure_time = one value for each TIFF file
            @unit
        exposure_period = one value for each TIFF file
            @unit
        exposure_number = integer component in the TIFF file name, for each TIFF file
        image_name = names of TIFF files
        count_cutoff = one value for each TIFF file
            @unit
    end_time = timestamp of the last image
    data:NXdata
        data --> /entry/instrument/detector/data

Version 0.1, Torsten Rathmann, Deutsches Klimarechenzentrum
"""
#################################################################################################################################

import nxs
import numpy
import Image
import sys
import os

class Error(Exception):
    """Exception raised for errors in the input.

    Attribute:
        msg  -- explanation of the error
    """

    def __init__(self, msg):
        print '\n*** Error: ', msg, '\n'


#################################### Main Part ###################################

#
# 1  Check Command Arguments
#

helptext = 'Converts directory of TIFF files to NeXus.\n' + \
           'Usage:   python tiff2nexus.py <inputdir> [ <outputfile> [hdf5|xml] ]\n'

argl = len(sys.argv)
if argl < 2 or argl > 4:
    print helptext
    sys.exit()
for i in range(1, argl):
    if sys.argv[i][0] == '-':
        print helptext
        sys.exit()
if argl > 3:
    outputformat = sys.argv[3]
    if outputformat not in ['hdf5','xml']:
        print helptext
        sys.exit()
else:
    outputformat = 'hdf5'

#
# 2  Check Input Directory
#

if not os.path.isdir(sys.argv[1]):
    raise IOError, sys.argv[1] + ' does not exist or is not a directory'
if not os.access(sys.argv[1], os.R_OK):
    raise IOError, 'Cannot read directory ' + sys.argv[1]
if not os.access(sys.argv[1], os.X_OK):
    raise IOError, 'Execute permission denied for directory ' + sys.argv[1]

#
# 3  Collect Input File Names (Names Ending with _#.tif, # = Integer Containing One Or More Digits)
#

filenames = sorted(os.listdir(sys.argv[1]))
nfiles = len(filenames)
inputfilenames = []
exposure_number = numpy.zeros((nfiles),'int32')
k = -1
for filename in filenames:
    if filename[-4:] != '.tif':
        print ' Non-TIFF file ' + filename + ' ignored'
        continue
    i = filename.rfind('_')
    if i < 0:
        print ' Non-Pilatus file ' + filename + ' ignored'
        continue
    try:
        j = int(filename[i+1:-4])
    except:
        print ' Non-Pilatus file ' + filename + ' ignored'
        continue
    k = k + 1  # inputfile index
    exposure_number[k] = j
    inputfilenames = inputfilenames + [filename]

n = k + 1  # number of input files
if n == 0:
    raise Error('No input files (files *_#.tif, # = integer containing one or more digits) found in directory ' + sys.argv[1])
if n < nfiles:
    exposure_number = exposure_number[:n]

#
# 4  Create NeXus File either as NeXus/HDF5 or as NeXus/XML
#

if argl > 2:
    outputfilename = sys.argv[2]
else:
    outputfilename = sys.argv[1] + '.h5'

if outputformat == 'hdf5':
    outputfile = nxs.open(outputfilename, mode='w5')
else:
    outputfile = nxs.open(outputfilename, mode='wx') # XML

#
# 5  Read and Validate Files and Write Data to NeXus
#

image_timestamps = ''
exposure_times   = numpy.zeros((n))
exposure_periods = numpy.zeros((n))
count_cutoffs    = numpy.zeros((n),'uint32')

for k in range(n):
#
# 5.1  read one input file
#
    path = sys.argv[1] + '/' + inputfilenames[k]
    inputfile = open(path) 
    content = inputfile.read()
    inputfile.close()
#
# 5.2  check file type
#
    if len(content) < 4096:     # The first 4096 bytes are the header of the input file
        raise Error('Header of file ' + inputfilenames[k] + ' is not complete')
    byte_order = content[0:2]                         # II for little endian
    tiff_id = ord(content[2]) + 2**8*ord(content[3])  # should be 42
    if not byte_order == 'II' or not tiff_id == 42:
        raise Error('Wrong type: ' + inputfilenames[k] + ' is not a Pilatus or MCCD file')
#
# 5.3  extract the text information from file header
#
    start_date  = content[30:49]
    detector_id = content[62:92]

    i = content.find('# Pixel_size') + 13
    j = content.find('# Trim_directory')
    j = j + content[j:4096].find('\r')
    try:
        pixel_size, sensor, exposure_time, exposure_period, tau, cutoff, threshold, n_excluded_pixels, excluded_pixels, flat_field, trim_directory = content[i:j].split('\r\n# ')
    except:
        raise Error(inputfilenames[k] + ' is not a valid Pilatus file: header text "# Pixel_size ... # Trim_directory" not valid')
#
# 5.4  get exposure time and unit
#
    try:
        dummy, expos_t, exposure_time_unit = exposure_time.split(' ')
        exposure_times[k] = float(expos_t)
    except:
        raise Error(inputfilenames[k] + ' is not a valid Pilatus file: exposure time information not valid')
#
# 5.5  get exposure period and unit
#
    try:
        dummy, expos_p, exposure_period_unit = exposure_period.split(' ')
        exposure_periods[k] = float(expos_p)
    except:
        raise Error(inputfilenames[k] + ' is not a valid Pilatus file: exposure period information not valid')
#
# 5.6  get count cutoff
#
    try:
        dummy, countcutoff, count_cutoff_unit = cutoff.split(' ')
        count_cutoffs[k] = int(countcutoff)
    except:
        raise Error(inputfilenames[k] + ' is not a valid Pilatus file: count cutoff information not valid')
#
# 5.7  process the TIFF header
#
    ifd_offset = ord(content[4]) + 2**8*ord(content[5]) + 2**16*ord(content[6]) + 2**24*ord(content[7])
    nr_entries = ord(content[ifd_offset]) + 2**8*ord(content[1+ifd_offset])
    for i in range(nr_entries):
        j = ifd_offset + i*12 + 14
        tag = ord(content[j]) + 2**8*ord(content[j+1])
        value_offset = ord(content[j+8]) + 2**8*ord(content[j+9]) + 2**16*ord(content[j+10]) + 2**24*ord(content[j+11])
        if tag == 257:
            height = value_offset
        elif tag == 256:
            width = value_offset
        elif tag == 258:
            bit_sample = value_offset
#
# 5.8  image size and bit-size of image data 
#
    try:
        size = width, height
    except:
        raise Error('Image size information missing in file ' + inputfilenames[k])
    try:
        if bit_sample == 16:
            mode = 'I;16'
        elif bit_sample == 32:
            mode = 'I'
        else:
            raise Error('Unexpected data mode information in file ' + inputfilenames[k] + ': neither int16 nor int32')
    except:
        raise Error('Data mode information missing in file ' + inputfilenames[k])
#
# 5.9  convert string into an image
#
    try:
        image = Image.fromstring(mode, size, content[4096:], "raw", mode, 0, 1)
    except:
        raise Error('Cannot read image data from file ' + inputfilenames[k])
#
# 5.10  convert image into an array of integers
#
    if image.mode == 'I;16':
	try:
            dataarray = numpy.array(image,'uint16')
        except:
	    raise Error(inputfilenames[k] + ': cannot write image data to uint16 array')
    else:
        try:
            dataarray = numpy.array(image,'uint32')
        except:
            raise Error(inputfilenames[k] + ': cannot write image data to uint32 array')
#
# 5.11  numpy organizes fast/slow axis differently, so swap
#
    dataarray = dataarray.transpose()
#
# 5.12  prepare dataarray for storing as 3-dimensional array
#
    dataarray = dataarray[numpy.newaxis, :]
#
# 5.13  read and write content from first file
#
    if k == 0:
        tif_version = content[110:129]
# 5.13.1  get pixel size and unit
        try:
            pixsize_x, pixel_unit_x, dummy, pixsize_y, pixel_unit_y = pixel_size.split(' ')
            pixel_size_x = float(pixsize_x)
            pixel_size_y = float(pixsize_y)
        except:
            raise Error(inputfilenames[k] + ' is not a valid Pilatus file: pixel size information not valid')
# 5.13.2  copy header information
        size0 = size
        mode0 = mode
        detector_id0 = detector_id
        pixel_size0 = pixel_size
        sensor0 = sensor
        tau0 = tau
        threshold0 = threshold
        n_excluded_pixels0 = n_excluded_pixels
        excluded_pixels0 = excluded_pixels
        flat_field0 = flat_field
        trim_directory0 = trim_directory
# 5.13.3  open group entry:NXentry
        outputfile.makegroup("entry", "NXentry")
        outputfile.opengroup("entry", "NXentry")
# 5.13.4  write title
        outputfile.makedata("title", dtype='char', shape=(len(sys.argv[1]),))
        outputfile.opendata("title")
        outputfile.putdata(sys.argv[1])
        outputfile.closedata()
# 5.13.5  write start time
        outputfile.makedata("start_time", dtype='char', shape=(len(start_date),))
        outputfile.opendata("start_time")
        outputfile.putdata(start_date)
        outputfile.closedata()
# 5.13.6  open group source:NXsource
        outputfile.makegroup("source", "NXsource")
        outputfile.opengroup("source", "NXsource")
# 5.13.7  write type
        outputfile.makedata("type", dtype='char', shape=(11,))
        outputfile.opendata("type")
        outputfile.putdata("Synchrotron")  # write default
        outputfile.closedata()
# 5.13.8  write name
        outputfile.makedata("name", dtype='char', shape=(8,))
        outputfile.opendata("name")
        outputfile.putdata("PetraIII")  # write default
        outputfile.closedata()
# 5.13.9  close group source:NXsource
        outputfile.closegroup()
# 5.13.10  open group instrument:NXinstrument
        outputfile.makegroup("instrument", "NXinstrument")
        outputfile.opengroup("instrument", "NXinstrument")
# 5.13.11  write name
        outputfile.makedata("name", dtype='char', shape=(10,))
        outputfile.opendata("name")
        outputfile.putdata("P03 GISAXS")  # write default
        outputfile.closedata()
# 5.13.12  open group detector:NXdetector
        outputfile.makegroup("detector", "NXdetector")
        outputfile.opengroup("detector", "NXdetector")
# 5.13.13  write aequatorial angle
        outputfile.makedata("aequatorial_angle", dtype='float64', shape=(1,))
        outputfile.opendata("aequatorial_angle")
        outputfile.putdata(0.)  # write default
        outputfile.closedata()
# 5.13.14  write azimuthal angle
        outputfile.makedata("azimuthal_angle", dtype='float64', shape=(1,))
        outputfile.opendata("azimuthal_angle")
        outputfile.putdata(0.)  # write default
        outputfile.closedata()
# 5.13.15  write beam center distance in x-direction
        outputfile.makedata("beam_center_x", dtype='float64', shape=(1,))
        outputfile.opendata("beam_center_x")
        outputfile.putdata(image.size[0] * pixel_size_x * 0.5)  # write default
        outputfile.putattr("unit", pixel_unit_x)
        outputfile.closedata()
# 5.13.16  write beam center distance in y-direction
        outputfile.makedata("beam_center_y", dtype='float64', shape=(1,))
        outputfile.opendata("beam_center_y")
        outputfile.putdata(image.size[1] * pixel_size_y * 0.5)  # write default
        outputfile.putattr("unit", pixel_unit_x)
        outputfile.closedata()
# 5.13.17  open array for image data
        outputfile.makedata("data", dtype=dataarray.dtype, shape=(nxs.UNLIMITED,image.size[0],image.size[1]))
        outputfile.opendata("data")
        outputfile.putattr("axes", "image:x:y")
#
# 5.14  compare header information of current input file with that of first file
#
    if size != size0:
        raise Error('Image data sizes differ in files ' + inputfilenames[0] + ' and ' + inputfilenames[k])
    if mode != mode0:
        raise Error('Image data types differ in files ' + inputfilenames[0] + ' and ' + inputfilenames[k])
    if detector_id != detector_id0:
        raise Error('Detector ID differs in files ' + inputfilenames[0] + ' and ' + inputfilenames[k])
    if pixel_size != pixel_size0:
        raise Error('Pixel size information differs in files ' + inputfilenames[0] + ' and ' + inputfilenames[k])
    if sensor != sensor0:
        raise Error('Sensor differs in files ' + inputfilenames[0] + ' and ' + inputfilenames[k])
    if tau != tau0:
        raise Error('Tau differs in files ' + inputfilenames[0] + ' and ' + inputfilenames[k])
    if threshold != threshold0:
        raise Error('Threshold_setting differs in files ' + inputfilenames[0] + ' and ' + inputfilenames[k])
    if n_excluded_pixels != n_excluded_pixels0:
        raise Error('N_excluded_pixels differ in files ' + inputfilenames[0] + ' and ' + inputfilenames[k])
    if excluded_pixels != excluded_pixels0:
        raise Error('Excluded_pixels differ in files ' + inputfilenames[0] + ' and ' + inputfilenames[k])
    if flat_field != flat_field0:
        raise Error('Flat_field differs in files ' + inputfilenames[0] + ' and ' + inputfilenames[k])
    if trim_directory != trim_directory0:
        raise Error('Trim_directory differs in files ' + inputfilenames[0] + ' and ' + inputfilenames[k])
#
# 5.15  write image data to NeXus
#
    outputfile.putslab(dataarray, (k,0,0), dataarray.shape)
#
# 5.16  append timestamp of current input file to string
#
    image_timestamps += '\n' + start_date

#
# 6  Get ID for Setting Link on Data Later
#

dataID = outputfile.getdataID()

#
# 7  Close Image Data in NeXus
#

outputfile.closedata()

#
# 8  Write Detector Distance
#

outputfile.makedata("distance", dtype='float64', shape=(1,))
outputfile.opendata("distance")
outputfile.putdata(0.)  # write default
outputfile.closedata()

#
# 9  Write Detector Polar Angle
#

outputfile.makedata("polar_angle", dtype='float64', shape=(1,))
outputfile.opendata("polar_angle")
outputfile.putdata(0.)  # write default
outputfile.closedata()

#
# 11  Write Detector Pixel Size in x-Direction
#

outputfile.makedata("x_pixel_size", dtype='float64', shape=(1,))
outputfile.opendata("x_pixel_size")
outputfile.putdata(pixel_size_x)
outputfile.putattr("unit", pixel_unit_x)
outputfile.closedata()

#
# 12  Write Detector Pixel Size in y-Direction
#

outputfile.makedata("y_pixel_size", dtype='float64', shape=(1,))
outputfile.opendata("y_pixel_size")
outputfile.putdata(pixel_size_y)
outputfile.putattr("unit", pixel_unit_y)
outputfile.closedata()

#
# 13  Open Group image_header:NXnote
#

outputfile.makegroup("image_header", "NXnote")
outputfile.opengroup("image_header", "NXnote")

#
# 14  Write Type
#

outputfile.makedata("type", dtype='char', shape=(10,))
outputfile.opendata("type")
outputfile.putdata("text/plain")
outputfile.closedata()

#
# 15  Write Description
#

d = detector_id + '\nconverted to NeXus from ' + tif_version.strip() + ' using tiff2nexus.py v 0.1\n' + sensor + '\n' + tau + '\n' + threshold + '\n' + n_excluded_pixels + '\n' + excluded_pixels + '\n' + flat_field + '\n' + trim_directory
outputfile.makedata("description", dtype='char', shape=(len(d),))
outputfile.opendata("description")
outputfile.putdata(d)
outputfile.closedata()

#
# 16  Close Group image_header:NXnote
#

outputfile.closegroup()

#
# 17  Close Group detector:NXdetector
#

outputfile.closegroup()

#
# 18  Open Group monochromator:NXmonochromator
#

outputfile.makegroup("monochromator", "NXmonochromator")
outputfile.opengroup("monochromator", "NXmonochromator")

#
# 19  Write Wavelength
#

outputfile.makedata("wavelength", dtype='float64', shape=(1,))
outputfile.opendata("wavelength")
outputfile.putdata(1.)  # write default
outputfile.closedata()

#
# 20  Write Wavelenth Spread
#

outputfile.makedata("wavelength_spread", dtype='float64', shape=(1,))
outputfile.opendata("wavelength_spread")
outputfile.putdata(0.)  # write default
outputfile.closedata()

#
# 21  Close Group monochromator:NXmonochromator
#

outputfile.closegroup()

#
# 22  Close Group instrument:NXinstrument
#

outputfile.closegroup()

#
# 23  Open Group sample:NXsample
#

outputfile.makegroup("sample", "NXsample")
outputfile.opengroup("sample", "NXsample")

#
# 24  write aequatorial angle
#

outputfile.makedata("aequatorial_angle", dtype='float64', shape=(1,))
outputfile.opendata("aequatorial_angle")
outputfile.putdata(0.)  # write default
outputfile.closedata()

#
# 25  Write Rotation Angle
#

outputfile.makedata("rotation_angle", dtype='float64', shape=(1,))
outputfile.opendata("rotation_angle")
outputfile.putdata(0.)  # write default
outputfile.closedata()

#
# 26  Write Sample Name
#

outputfile.makedata("name", dtype='char', shape=(len(sys.argv[1]),))
outputfile.opendata("name")
outputfile.putdata(sys.argv[1])  # input directory name
outputfile.closedata()

#
# 27  Write String Containing Timestamps
#

outputfile.makedata("image_timestamp", dtype='char', shape=(len(image_timestamps),))
outputfile.opendata("image_timestamp")
outputfile.putdata(image_timestamps)
outputfile.closedata()

#
# 28  Write Exposure Times
#

outputfile.makedata("exposure_time", dtype=exposure_times.dtype, shape=exposure_times.shape)
outputfile.opendata("exposure_time")
outputfile.putdata(exposure_times)
outputfile.putattr("unit", exposure_time_unit)
outputfile.closedata()

#
# 29  Write Exposure Periods
#

outputfile.makedata("exposure_period", dtype=exposure_periods.dtype, shape=exposure_periods.shape)
outputfile.opendata("exposure_period")
outputfile.putdata(exposure_periods)
outputfile.putattr("unit", exposure_period_unit)
outputfile.closedata()

#
# 30  Write Exposure Number
#

outputfile.makedata("exposure_number", dtype=exposure_number.dtype, shape=exposure_number.shape)
outputfile.opendata("exposure_number")
outputfile.putdata(exposure_number)
outputfile.closedata()

#
# 31  Write Input File Names
#

filenames = inputfilenames[0]
for k in range(1, n):
   filenames = filenames + '\n' + inputfilenames[k]
outputfile.makedata("image_name", dtype='char', shape=(len(filenames),))
outputfile.opendata("image_name")
outputfile.putdata(filenames)
outputfile.closedata()

#
# 32  Write Count Cutoffs
#

outputfile.makedata("count_cutoff", dtype=count_cutoffs.dtype, shape=count_cutoffs.shape)
outputfile.opendata("count_cutoff")
outputfile.putdata(count_cutoffs)
outputfile.putattr("unit", count_cutoff_unit)
outputfile.closedata()

#
# 33  Close Group sample:NXsample
#

outputfile.closegroup()

#
# 34  Write End Time to Group NXentry
#

outputfile.makedata("end_time", dtype='char', shape=(len(start_date),))
outputfile.opendata("end_time")
outputfile.putdata(start_date)
outputfile.closedata()

#
# 35  Make Link to Image Data
#

outputfile.makegroup("data", "NXdata")
outputfile.opengroup("data", "NXdata")
outputfile.makelink(dataID)
outputfile.closegroup()

#
# 36  Close Group entry:NXentry
#

outputfile.closegroup()

#
# 37  Close NeXus File
#

outputfile.close()
